;created as an attempt to replicate results found by Cremers as expressed in Key Exchange in IPsec Revisited.
;(Or rather from input of ikev2-child.cpp)
(herald "IKEv2 Child" (algebra diffie-hellman) (goals-sat) (bound 8) (limit 500))

(defmacro (SK_e) (hash I R secret "e"));initiator-encryption
(defmacro (SK_ee) (hash R I secret "e"));responder-encryption
(defmacro (SK_d) (hash I R secret "d"));initiator-derrivation
(defmacro (SK_dd) (hash R I secret "d"));responder-derrivation

;Macros to reflect what certian parties do not know
(defmacro (zi) (exp Gr i))
(defmacro (zr) (exp Gi r))

(defmacro (SKi) (hash "KDF" (SK_d) (zi) ni nr))
(defmacro (SKii) (hash "KDF" (SK_dd) (zi) ni nr))
(defmacro (SKr) (hash "KDF" (SK_d) (zr) ni nr))
(defmacro (SKrr) (hash "KDF" (SK_dd) (zr) ni nr))

(defmacro (ilength) 2)
(defmacro (rlength) 1)

(defmacro (SKf) (hash "KDF" (SK_d) (exp (gen) (mul i r)) ni nr))
(defmacro (SKff) (hash "KDF" (SK_dd) (exp (gen) (mul i r)) ni nr))

;(defmacro (SKir) (hash "KDF" (ltk I R) ni nr))
;(defmacro (SKri) (hash "KDF" (ltk R I) ni nr))

(defprotocol child diffie-hellman
	     (defrole init
		      (vars (ni nr lst3 algo3 secret text) (i expn) (Gr base) (I R name))
		      (trace
			(send (cat I R (enc lst3 ni (exp (gen) i) (SK_e))))
			(recv (cat R I (enc algo3 nr Gr (SK_ee))))
			(send "blab"))
		      (uniq-orig ni)
		      (uniq-gen i)
		      (neq (Gr (gen)) (lst3 algo3)))
	     (defrole resp
		      (vars (ni nr lst3 algo3 secret text) (r expn) (Gi base) (I R name))
		      (trace
			(recv (cat I R (enc lst3 ni Gi (SK_e))))
			(send (cat R I (enc algo3 nr (exp (gen) r) (SK_ee)))))
		      (uniq-orig nr)
		      (uniq-gen r)
		      (neq (Gi (gen)))
		      (priority (0 4)))
	     (defrole blab-secret
               (vars (secret text))
               (trace
                (recv "blab")
                (send secret))
               (uniq-orig secret)))


;Weakest - no keys are leaked. 
 (defskeleton child
	     (vars (ni nr lst3 algo3 secret text) (i expn) (Gr base) (I R name))
	     (defstrand init 2 (I I) (R R) (i i) (Gr Gr) (secret secret))
	     (deflistener (SKi))
	     (non-orig secret))

(defskeleton child
	     (vars (ni nr lst3 algo3 secret text) (i expn) (Gr base) (I R name))
	     (defstrand init 2 (I I) (R R) (i i) (Gr Gr) (secret secret))
             (deflistener (SKii))
	     (non-orig secret))

(defskeleton child
	     (vars (ni nr lst3 algo3 secret text) (r expn) (Gi base) (I R name))
	     (defstrand resp 2 (I I) (R R) (r r) (Gi Gi) (secret secret))
             (deflistener (SKr))
	     (non-orig secret))

(defskeleton child
	     (vars (ni nr lst3 algo3 secret text) (r expn) (Gi base) (I R name))
	     (defstrand resp 2 (I I) (R R) (r r) (Gi Gi) (secret secret))
             (deflistener (SKrr))
	     (non-orig secret))


; Weak Forward Secrecy
(defskeleton child
  (vars (ni nr lst3 algo3 secret text) (I R name) (r i expn))
  (defstrand init 3 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gr (exp (gen) r)) (i i) (secret secret))
  (defstrand resp 2 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gi (exp (gen) i)) (r r) (secret secret))
  (defstrand blab-secret 2 (secret secret))
  (deflistener (SKf))
  (precedes ((0 0) (1 0)) ((1 1) (0 1)) ((0 2) (2 0)))
  (ind-zero-in (r (exp (gen) i)))
  (neq ((exp (gen) i) (gen)) ((exp (gen) r) (gen)))
  (uniq-gen r i)
  (uniq-orig ni nr))

(defskeleton child
  (vars (ni nr lst3 algo3 secret text) (I R name) (r i expn))
  (defstrand init 3 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gr (exp (gen) r)) (i i) (secret secret))
  (defstrand resp 2 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gi (exp (gen) i)) (r r) (secret secret))
  (defstrand blab-secret 2 (secret secret))
  (deflistener (SKff))
  (precedes ((0 0) (1 0)) ((1 1) (0 1)) ((0 2) (2 0)))
  (ind-zero-in (r (exp (gen) i)))
  (neq ((exp (gen) i) (gen)) ((exp (gen) r) (gen)))
  (uniq-gen r i)
  (uniq-orig ni nr))


;Perfect forward secrecy
(defskeleton child
  (vars (ni nr lst3 algo3 secret text) (I R name) (Gr base) (r i expn))
  (defstrand init 3 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gr Gr) (i i) (secret secret))
     (defstrand blab-secret 2 (secret secret))
     (deflistener (SKi))
     (precedes ((0 2) (1 0))))

(defskeleton child
  (vars (ni nr lst3 algo3 text) (I R name) (i expn) (Gr base) (secret text))
  (defstrand init 3 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gr Gr) (i i) (secret secret))
     (defstrand blab-secret 2 (secret secret))
     (deflistener (SKii))
     (precedes ((0 2) (1 0))))

(defskeleton child
  (vars (ni nr lst3 algo3 secret text) (I R name) (r expn) (Gi base))
  (defstrand resp 2 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gi Gi) (r r) (secret secret))
     (defstrand blab-secret 2 (secret secret))
     (deflistener (SKr))
     (precedes ((0 1) (1 0))))

(defskeleton child
  (vars (ni nr lst3 algo3 secret text) (I R name) (r expn) (Gi base))
  (defstrand resp 2 (ni ni) (nr nr) (lst3 lst3) (algo3 algo3) (I I)
    (R R) (Gi Gi) (r r) (secret secret))
     (defstrand blab-secret 2 (secret secret))
     (deflistener (SKrr))
     (precedes ((0 1) (1 0))))

(defgoal child
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr lst3 algo3 secret text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "algo3" iend algo3) (p "init" "lst3" iend lst3)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "secret" iend secret)
		(non secret))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R))))))
  ; Weak agreement - my peer and I agree on our names
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr lst3 algo3 secret text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "algo3" iend algo3) (p "init" "lst3" iend lst3)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "secret" iend secret)
		(non secret))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R) (p "init" "R" i1 I))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr lst3 algo3 secret text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "algo3" iend algo3) (p "init" "lst3" iend lst3)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "secret" iend secret)
		(non secret))
	   (exists ((r1 node))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			))))

  )(comment
  ; Full agreement - my peer and I agree on the full transcript
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr lst3 algo3 secret text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "algo3" iend algo3) (p "init" "lst3" iend lst3)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "secret" iend secret)
		(non secret))
	   (exists ((r1 node) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "ni" r1 ni) (p "resp" "nr" r1 nr)
			(p "resp" "r" r1 r) (p "resp" "Gi" r1 (exp (gen) i))
			(p "resp" "lst3" r1 lst3) (p "resp" "algo3" r1 algo3)
			(p "resp" "secret" r1 secret)))))
  )

(defgoal child
  ; Aliveness - my peer has taken an action
  (forall ((rend node) (I R name) (ni nr lst3 algo3 secret text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "secret" rend secret)
		(p "resp" "lst3" rend lst3) (p "resp" "algo3" rend algo3) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi) 
		(non secret))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((rend node) (I R name) (ni nr lst3 algo3 secret text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "secret" rend secret)
		(p "resp" "lst3" rend lst3) (p "resp" "algo3" rend algo3) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi) 
		(non secret))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I) (p "resp" "I" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I) (p "init" "R" i1 R))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((rend node) (I R name) (ni nr lst3 algo3 secret text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "secret" rend secret)
		(p "resp" "lst3" rend lst3) (p "resp" "algo3" rend algo3) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi) 
		(non secret))
	   (exists ((i1 node))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
			))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((rend node) (I R name) (ni nr lst3 algo3 secret text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "secret" rend secret)
		(p "resp" "lst3" rend lst3) (p "resp" "algo3" rend algo3) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi) 
		(non secret))
	   (exists ((i1 node) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
			(p "init" "ni" i1 ni) (p "init" "nr" i1 nr)
			(p "init" "algo3" i1 algo3) (p "init" "lst3" i1 lst3)
			(p "init" "secret" i1 secret) 
		        (p "init" "Gr" i1 (exp (gen) r)) (p "init" "i" i1 i)
			))))
  (comment "Responder point of view; agreement / weak agreement / aliveness goals")
  )
